package Plugins::MSAD;
use Config::General;
use MIME::Base64;

my(@f_attribs, $att, $val, $f_attrib);

my $conf = new Config::General(
    -ConfigFile      => $config::plugins_dir . '/MSAD.conf',
    -InterPolateVars => TRUE,
    -AutoLaunder     => TRUE
);
my %cfg = $conf->getall;

my $VERSION     = 1.0;
my $plugin_name = 'MSAD';
my $this_pkg    = (caller(0))[0] . '::' . $plugin_name;

$Plugins::MSAD::panic = $cfg{'panic_on_errors'};
@Plugins::MSAD::AD_admins = keys %{ $cfg{'admin_users'} };
@Plugins::MSAD::AD_exclude = keys %{ $cfg{'exclude_users'} };
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': getting admin list: ' . join(", ", @Plugins::MSAD::AD_admins)) if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': getting exclusion list: ' . join(", ", @Plugins::MSAD::AD_exclude)) if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': users disabled=' . $cfg{'AD_new_users_disabled'})            if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': users password never expires=' . $cfg{'AD_new_users_pne'})   if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': users password expires now=' . $cfg{'AD_new_users_pwd_exp'}) if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': use custom suffix=' . $cfg{'use_custom_suffix'})             if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': custom suffix=' . $cfg{'AD_custom_suffix'})                  if $config::aosi_debug_level > 1;
AOSICFG::write_ws_log('-=- ' . $this_pkg . ': panic on errors=' . $Plugins::MSAD::panic)                   if $config::aosi_debug_level > 1;

$Plugins::MSAD::filters = {};
if ($cfg{'custom_filter'}) {
    AOSICFG::write_ws_log('-=- ' . $this_pkg . ': custom filter=' . $cfg{'custom_filter'}) if $config::aosi_debug_level > 1;
    @f_attribs = split(/[,;]/, $cfg{'custom_filter'});
    foreach $f_attrib (@f_attribs) {
        ($att, $val) = split(/=/, $f_attrib);
        if (not exists $Plugins::MSAD::filters->{$att}) {
            $Plugins::MSAD::filters->{$att} = [];
        }
        push(@{ $Plugins::MSAD::filters->{$att} }, $val);
    }
    AOSICFG::write_ws_log('-=- ' . $this_pkg . ': filters=' . AOSICFG::dump_hash($Plugins::MSAD::filters)) if $config::aosi_debug_level > 1;
}

if($cfg{'AD_profile_path'}){
    AOSICFG::write_ws_log('-=- ' . $this_pkg . ': profile path=' . $cfg{'AD_profile_path'})                  if $config::aosi_debug_level > 1;
}


######################################################################
# Check if an uid is on the list of admins.
# Returns: a number of hits (usualy 0 or 1)
######################################################################
sub _user_is_in_list {
    my ($in, $list) = @_;
    my ($ok, $realm, $uid, $user, @users, $this_fun);
    $this_fun = (caller(0))[3];

    $ok = 0;
    ($uid, $realm) = AOSICFG::user_split($in);
    foreach $user (@$list) {
        chomp($user);
        if ($user eq $uid) {
            $ok++;
        }
    }
    AOSICFG::write_ws_log('|> ' . $this_fun . ': user ' . $in . ' is in list ['.join(", ", @$list).']: ' . $ok) if $config::aosi_debug_level > 2;
    return $ok;
}

######################################################################
# Wrapper for _user_is_ADadmin
######################################################################
sub user_is_ADadmin {
    my ($in) = @_;
    return _user_is_in_list($in, \@Plugins::MSAD::AD_admins);
}

######################################################################
# Check if an uid is in exclusion list
######################################################################
sub user_is_excluded {
    my ($in) = @_;

    return _user_is_in_list($in, \@Plugins::MSAD::AD_exclude);
}


######################################################################
# Get the full AD DN for an uid (sAMAccountName)
# Retuns: DN (if found), error code and error message
######################################################################
sub get_ADDN {
    my ($ldap, $base, $uid) = @_;
    my ($code, $error, @ents, $req_dn, $log, $this_fun);
    $this_fun = (caller(0))[3];

    AOSICFG::write_ws_log("|> $this_fun(in): base=" . $base . ";uid=" . $uid) if $config::aosi_debug_level > 3;
    my $search = $ldap->search(
        base   => $base,
        filter => '(sAMAccountName=' . $uid . ')',
        attrs  => ['distinguishedName'],
        scope  => 'sub'
    );
    $code  = $search->code;
    $error = $search->error;
    $error =~ s/[\x00\n\r]//g;
    AOSICFG::write_ws_log("|> $this_fun(search result): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 3;

    if ($code == 0) {
        @ents = $search->entries;
        if ($#ents > -1) {
            $req_dn = $ents[0]->get('distinguishedName')->[0];
        }
        else {
            $code  = -10009;
            $error = 'User ' . $uid . ' does not exist!';
        }
    }

    AOSICFG::write_ws_log("|> $this_fun(out): req_dn=" . $req_dn . ";code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 3;
    return $req_dn, $code, $error;
}

######################################################################
# Extract the uid part from DN
######################################################################
sub get_uid {
    my ($dn) = @_;
    my ($uid, $tmp);

    ($uid, $tmp) = split(/,/, $dn);
    ($tmp, $uid) = split(/=/, $uid);
    return $uid;
}

######################################################################
# Convert a base DN to domain
# (eg. DC=intranet,DC=domain,DC=hr -> intranet.domain.hr)
######################################################################
sub basedn2domain {
    my ($basedn) = @_;
    my ($out, @temp);

    $out = $basedn;
    $out =~ s/DC=//ig;
    @temp = split(/,/, $out);
    $out = join(".", @temp);
    return $out;
}

##################################################################
# Check filter
##################################################################
sub ck_filter {
    my ($ldap, $base_dn, $entry_in, $user, $pwd, $new, $filters) = @_;
    my ($this_fun, $code,  $msg,    $attrs, $attr,  $filter, $a_name, $a_vals,   $f_vals,   $ok);
    my ($a_val,    $f_val, @f_keys, $mesg,  $mesg2, $dn,     $uid,    @f_search, $s_filter, %options);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;
    $ok       = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): base=$base_dn;user=$user;entry_in=$entry_in;new=$new;") if $config::aosi_debug_level > 1;
    @f_keys = keys %$filters;
    if ($#f_keys > -1) {
        if ($new == 1) {    # New user - check data in XML
            $attrs = $entry_in->{'attrs'};
            AOSICFG::write_ws_log("|> $this_fun(new): attrs=" . join(",", @$attrs) . ";") if $config::aosi_debug_level > 2;
            foreach $attr (@$attrs) {
                $a_name = $attr->{'att'};
                $a_vals = $attr->{'val'};
                AOSICFG::write_ws_log("|> $this_fun: att=$a_name") if $config::aosi_debug_level > 3;
                foreach $a_val (@$a_vals) {
                    AOSICFG::write_ws_log("|> $this_fun: val=$a_val") if $config::aosi_debug_level > 3;
                    foreach $filter (keys %$filters) {
                        $f_vals = $filters->{$filter};
                        AOSICFG::write_ws_log("|> $this_fun: f_name=$filter;") if $config::aosi_debug_level > 3;
                        foreach $f_val (@$f_vals) {
                            AOSICFG::write_ws_log("|> $this_fun: f_val=$f_val;") if $config::aosi_debug_level > 3;
                            if (($a_name eq $filter) and ($a_val eq $f_val)) {
                                $ok++;
                            }
                        }
                    }
                }
            }
            AOSICFG::write_ws_log("|> $this_fun: ok=$ok;") if $config::aosi_debug_level > 1;
            if ($ok > 0) {
                $code = 0;
            }
            else {
                $code = -10060;
                $msg  = "User attributes do not meet the filter (" . $cfg{'custom_filter'} . ") requirements!";
            }
        }
        else {    # Old user - check data in LDAP
            $dn       = $entry_in->{'dn'};
            $uid      = get_uid($dn);
            @f_search = ();
            foreach $filter (keys %$filters) {
                $f_vals = $filters->{$filter};
                foreach $f_val (@$f_vals) {
                    push(@f_search, $filter . '=' . $f_val);
                }
            }
            $s_filter = '(uid=' . $uid . ')';
            AOSICFG::write_ws_log("|> $this_fun: search filter1=$s_filter;") if $config::aosi_debug_level > 2;
            %options = (
                base      => $base_dn,
                filter    => $s_filter,
                attrs     => ['dn'],
                timelimit => 180
            );
            $mesg = $ldap->search(%options);
            $code = $mesg->code;
            $msg  = $mesg->error;
            $msg =~ s/[\x00\n\r]//g;
            AOSICFG::write_ws_log("|> $this_fun(looking for uid): code=$code;msg=$msg;entries=" . $#{ $mesg->entries }) if $config::aosi_debug_level > 2;

            if (($code == 0) and ($#{ $mesg->entries } > -1)) {
                $s_filter = '(&(uid=' . $uid . ')(' . join(")(", @f_search) . '))';
                AOSICFG::write_ws_log("|> $this_fun: search filter2=$s_filter;") if $config::aosi_debug_level > 2;
                %options = (
                    base      => $base_dn,
                    filter    => $s_filter,
                    attrs     => ['dn'],
                    timelimit => 180
                );
                $mesg2 = $ldap->search(%options);
                $code  = $mesg2->code;
                $msg   = $mesg2->error;
                $msg =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun: code=$code;msg=$msg;") if $config::aosi_debug_level > 3;
                if ($code == 0) {
                    AOSICFG::write_ws_log("|> $this_fun: entries=" . $#{ $mesg2->entries }) if $config::aosi_debug_level > 3;
                    if ($#{ $mesg2->entries } == -1) {
                        $code = -10061;
                        $msg  = "User attributes do not meet the filter (" . $cfg{'custom_filter'} . ") requirements!";
                    }
                }
            }
        }
    }
    else {
        AOSICFG::write_ws_log("|> $this_fun: no filter") if $config::aosi_debug_level > 1;
    }
    AOSICFG::write_ws_log("|> $this_fun(out): code=$code;msg=$msg") if $config::aosi_debug_level > 1;
    return $code, $msg;
}

######################################################################
# Connect to AD, first as AOSIRO account to find out the realm
# AD DN for a user and then as the user
# Returns: AD LDAP object, error code and error message
######################################################################
sub AD_RO_connect {
    my ($user, $pwd, $check_admin) = @_;
    my ($ad_ldap, $mesg_ro, $code_ro, $error_ro, $req_dn, $code, $msg, $error, $admin);
    my ($dec_pwd, $req, $uid, $realm, $this_fun, $out_panic);

    $out_panic = 0;
    $this_fun  = (caller(0))[3];
    ($uid, $realm) = split(/\@/, $user);
    AOSICFG::write_ws_log("|> $this_fun: AD at " . $cfg{'AD_hostname'} . ":" . $cfg{'AD_port'}) if $config::aosi_debug_level > 2;
    $ad_ldap = Net::LDAP->new('ldaps://' . $cfg{'AD_hostname'}, port => $cfg{'AD_port'}, timeout => 20);
    if ($ad_ldap) {
        $mesg_ro  = $ad_ldap->bind($cfg{'AD_aosiro_dn'}, password => $cfg{'AD_aosiro_pwd'});
        $code_ro  = $mesg_ro->code;
        $error_ro = $mesg_ro->error;
        $error_ro =~ s/[\x00\n\r]//g;
        AOSICFG::write_ws_log("|> $this_fun: code=$code_ro;error=$error_ro;") if $config::aosi_debug_level > 2;
        if ($code_ro == 0) {
            ($req_dn, $code, $error) = get_ADDN($ad_ldap, $cfg{'AD_base'}, $uid);
            AOSICFG::write_ws_log("|> $this_fun(AOSIRO search): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
            if ($code == 0) {
                if($check_admin == 1){
                    $admin = user_is_ADadmin($user);
                    if ($admin > 0) {
                        AOSICFG::write_ws_log("|> $this_fun(req DN): " . $req_dn);
                        $dec_pwd = MIME::Base64::decode_base64($pwd);
                        $req     = $ad_ldap->bind($req_dn, password => $dec_pwd);
                        $code    = $req->code;
                        $error   = $req->error;
                        $error =~ s/[\x00\n\r]//g;
                        AOSICFG::write_ws_log("|> $this_fun(req bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
                        $msg = $error;
                    }
                    else {
                        $msg  = 'User ' . $uid . ' has no admin privileges!';
                        $code = -10003;
                    }
                }
            }
            else {
                $msg  = 'AOSIRO user could not Search for ' . $uid . ' (AD code: ' . $code . '; AD message: ' . $error . ').';
                $code = -10002;
                $out_panic = $Plugins::MSAD::panic;
            }
        }
        else {
            $msg  = 'AOSIRO user could not bind to AD (AD code: ' . $code_ro . '; AD message: ' . $error_ro . ').';
            $code = -10001;
            $out_panic = $Plugins::MSAD::panic;
        }
    }
    else {
        $code = -10000;
        $msg  = "Connection to MS AD server " . $cfg{'AD_hostname'} . ":" . $cfg{'AD_port'} . " failed: $@";
        $out_panic = $Plugins::MSAD::panic;
    }
    return $ad_ldap, $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeAddUser {
    my ($status_hash, $ldap, $base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $ad_ldap, $mesg_ro,  $code_ro,   $error_ro, $key);
    my ($req_dn,   $code,    $msg,      $error,     $req,      $dec_pwd, $uid, $realm, $search);
    my ($new_uid,  $new_dn,  $code_new, $error_new, $temp,     @entries, $admin, $panic, $out_panic);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
    foreach $key (keys %$status_hash) {
        $panic = $status_hash->{$key}->{'before'}->{'panic'};
        AOSICFG::write_ws_log($this_fun . '(before hash): ' . $key . '; panic=' . $panic . ';') if $config::aosi_debug_level > 2;
    }

    ($code, $msg) = ck_filter($ldap, $base_dn, $entry_in, $user, $pwd, 1, $Plugins::MSAD::filters);
    if ($code == 0) {
        ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1);
        if ($code == 0) {
            $new_dn = $entry_in->{'dn'};
            $new_uid = get_uid($new_dn);
            if(user_is_excluded($new_uid)){
                $code = -10011;
                $msg = 'Uid '.$new_uid.' is on exclusion list!';
                $out_panic = $Plugins::MSAD::panic;
            }
            else {
                $search = $ad_ldap->search(
                    base   => $cfg{'AD_base'},
                    filter => '(sAMAccountName=' . $new_uid . ')',
                    attrs  => ['distinguishedName'],
                    scope  => 'sub'
                );
                $code_new  = $search->code;
                $error_new = $search->error;
                $error_new =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun(looking for): uid=" . $new_uid . ";code=" . $code_new . ";error=" . $error_new)
                  if $config::aosi_debug_level > 2;
                if ($code_new == 0) {
                    @entries = $search->entries;
                    AOSICFG::write_ws_log("|> $this_fun(user exists): uid=" . $new_uid . ";#entries=" . $#entries) if $config::aosi_debug_level > 2;
                    if ($#entries > -1) {
                        $code      = -10010;
                        $msg       = "User $new_uid allready exists!";
                        $out_panic = $Plugins::MSAD::panic;
                    }
                    else {
                        $ent_attrs   = $entry_in->{'attrs'};
                        AOSICFG::write_ws_log("|> $this_fun: ent_attrs=[" .join(", ", @$ent_attrs). "]") if $config::aosi_debug_level > 2;
                        foreach $attr(@$ent_attrs){
                            $att_name = $attr->{'att'};
                            AOSICFG::write_ws_log("|> $this_fun: att=$att_name") if $config::aosi_debug_level > 2;
                            if($att_name eq 'userPassword'){
                                $msg   = encode_base64($attr->{'val'}->[0]);
                            }
                        }
                    }
                }
                else {
                    $msg = '';
                }
            }
        }
    }

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Add an account
##################################################################
sub afterAddUser {
    my ($status_hash, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $ad_ldap, $mesg_ro,  $code_ro,   $error_ro,  $uid,         $realm);
    my ($req_dn,   $code,    $msg,      $error,     $req,       $dec_pwd,     $search);
    my ($new_uid,  $new_dn,  $code_new, $error_new, $temp,      $admin,       $attr);
    my ($dn,       $add_dn,  $attrs,    $ent_attrs, $out_panic, $before_code, $before_msg, $before_panic);
    my ($PWD,      $uqid,    $sn,       $givenName, $mesg_add,  $uac,         $mesg_mod);
    my ($msg1,     $error1,  $code1,    $npwd,      $msg2,      $error2,      $code2, $mesg_pwd, $orig_pwd, $mail, $path);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;
    ($uid, $realm) = split(/\@/, $user);

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
    $before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
    $before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};

    if ($before_code == 0 and $before_panic == 0) {
        $orig_pwd = decode_base64($before_msg);
        if ($code == 0) {    # This one is from parse_ldap_xml
            ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1);
            if ($code == 0) {
                my $mesg = $ad_ldap->bind($cfg{'AD_aosiwrt_dn'}, password => $cfg{'AD_aosiwrt_pwd'});
                $code  = $mesg->code;
                $error = $mesg->error;
                $error =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun(AOSIWRT bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
                if ($code == 0) {
                    $dn     = $entry_in->{'dn'};
                    $uid    = get_uid($dn);
                    $add_dn = 'CN=' . $uid . ',' . $cfg{'AD_new_dn'} . ',' . $cfg{'AD_base'};
                    AOSICFG::write_ws_log("|> $this_fun: add_dn=" . $add_dn) if $config::aosi_debug_level > 1;
                    $ent_attrs = $entry_in->{'attrs'};
                    $attrs     = [];
                    foreach $attr (@$ent_attrs) {    # Fetch some attributes
                        if (($attr->{'att'} eq 'userPassword')) {
                            #$PWD = join(",", @{ $attr->{'val'} });
                            $PWD = $orig_pwd;
                        }
                        elsif (($attr->{'att'} eq 'hrEduPersonUniqueID')) {
                            $uqid = join(",", @{ $attr->{'val'} });
                        }
                        elsif (($attr->{'att'} eq 'sn')) {
                            $sn = join(" ", @{ $attr->{'val'} });
                        }
                        elsif (($attr->{'att'} eq 'mail')) {
                            $mail = join(" ", @{ $attr->{'val'} });
                        }
                        elsif (($attr->{'att'} eq 'givenName')) {
                            $givenName = join(" ", @{ $attr->{'val'} });
                        }
                        else {
                            if ($attr->{'att'} !~ /^hrEduPerson/) {
                                #push(@$attrs, $attr->{'att'} => $attr->{'val'});
                            }
                        }
                    }

                    # Composing AD attributes
                    push(@$attrs, 'cn'             => [$uid]);
                    push(@$attrs, 'sAMAccountName' => [$uid]);
                    push(@$attrs, 'uid'            => [$uid]);
                    if ($cfg{'use_custom_suffix'} == 1) {
                        push(@$attrs, 'userPrincipalName' => [ $uid . '@' . $cfg{'AD_custom_suffix'} ]);
                    }
                    else {
                        push(@$attrs, 'userPrincipalName' => [ $uid . '@' . basedn2domain($cfg{'AD_base'}) ]);
                    }
                    push(@$attrs, 'displayName' => [ $givenName . ' ' . $sn ]);
                    push(@$attrs, 'description' => [ 'hrEduPersonUniqueID: ' . $uqid ]);
                    push(@$attrs, 'sn'          => [$sn]);
                    push(@$attrs, 'givenName'   => [$givenName]);
                    if($mail ne ''){
                        push(@$attrs, 'mail'        => [$mail]);
                    }
                    $path = $cfg{'AD_profile_path'};
                    if ($path ne '') {

                        if (rindex($path, '\\') != (length($path) - 1)) {    # Fix the path
                            $path .= '\\';
                        }
                        $path .= $uid;                                       # Append sAMAccountName to path
                        push(@$attrs, 'profilePath' => [$path]);
                    }
                    push(@$attrs, 'objectClass' => [ 'top', 'person', 'user', 'organizationalPerson' ]);
                    AOSICFG::write_ws_log("|> $this_fun: attrs=" . AOSICFG::dump_array($attrs)) if $config::aosi_debug_level > 3;
                    $mesg_add = $ad_ldap->add($add_dn, attrs => $attrs);
                    $code     = $mesg_add->code;
                    $error    = $mesg_add->error;
                    $error =~ s/[\x00\n\r]//g;
                    AOSICFG::write_ws_log("|> $this_fun(adding): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;

                    # UserAccountControl flags
                    # ACCOUNTDISABLE        0x0002      2
                    # PASSWD_NOTREQD        0x0020      32
                    # PASSWD_CANT_CHANGE    0x0040      64
                    # NORMAL_ACCOUNT        0x0200      512
                    # DONT_EXPIRE_PASSWORD  0x10000     65536
                    # PASSWORD_EXPIRED      0x800000    8388608
                    if ($code == 0) {
                        $uac = 512;
                        if ($cfg{'AD_new_users_disabled'} == 1) {
                            $uac += 2;    # Set account is disabled
                        }
                        if ($cfg{'AD_new_users_pne'} == 1) {
                            $uac += 65536;    # Set password never expires
                        }
                        if ($cfg{'AD_new_users_pwd_exp'} == 1) {
                            $uac += 8388608;    # Set password expired
                        }
                        AOSICFG::write_ws_log("|> $this_fun(uac): $uac") if $config::aosi_debug_level > 2;
                        $mesg_mod = $ad_ldap->modify($add_dn, replace => { 'userAccountControl' => $uac });
                        $code1    = $mesg_mod->code;
                        $error1   = $mesg_mod->error;
                        $error1 =~ s/[\x00\n\r]//g;
                        AOSICFG::write_ws_log("|> $this_fun(adding user): code=$code1;error=$error1") if $config::aosi_debug_level > 2;
                        $msg1 = 'User ' . $uid . ' added but could not set properties (AD code: ' . $code1 . '; AD message: ' . $error1 . ').';

                        # Set password
                        map { $npwd .= "$_\000" } split(//, "\"$PWD\"");    # Encode to UTF-16
                        $mesg_pwd = $ad_ldap->modify($add_dn, replace => { 'unicodePwd' => $npwd });
                        $code2    = $mesg_pwd->code;
                        $error2   = $mesg_pwd->error;
                        $error2 =~ s/[\x00\n\r]//g;
                        AOSICFG::write_ws_log("|> $this_fun(setting pwd): code=$code2;error=$error2") if $config::aosi_debug_level > 2;
                        $msg2 = 'User ' . $uid . ' added but could not set password (AD code: ' . $code2 . '; AD message: ' . $error2 . ').';

                        $code = $code1 + $code2;
                        if ($code == 0) {
                            $msg  = 'User ' . $uid . ' added (AD code: ' . $code . '; AD message: ' . $error . ').';
                            $code = 0;
                        }
                        else {
                            $msg  = $msg1 . "\n" . $msg2;
                            $code = -10023;
                        }
                    }
                    else {
                        $msg  = 'User ' . $uid . ' could not be added (AD code: ' . $code . '; AD message: ' . $error . ').';
                        $code = -10022;
                    }
                }
                else {
                    $msg  = 'AOSIWRT user could not bind to AD (AD code: ' . $code . '; AD message: ' . $error . ').';
                    $code = -10021;
                }
            }
        }
    }
    else {
        $code = -10020;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }
    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
# Check if an account exists
##################################################################
sub beforeDeleteUser {
    my ($status_hash, $ldap, $base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $ad_ldap, $mesg_ro,  $code_ro,   $error_ro, $msg,     $search);
    my ($req_dn,   $code,    $error,    $req,       $dec_pwd,  $uid,     $realm);
    my ($del_uid,  $del_dn,  $code_del, $error_del, $temp,     @entries, $out_panic);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;
    ($uid, $realm) = split(/\@/, $user);

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    ($code, $msg) = ck_filter($ldap, $base_dn, $entry_in, $user, $pwd, 0, $Plugins::MSAD::filters);
    if ($code == 0) {
        ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1);
        if ($code == 0) {
            $del_dn  = $entry_in->{'dn'};
            $del_uid = get_uid($del_dn);
            if(user_is_excluded($del_uid)){
                $code = -10032;
                $msg = 'Uid '.$del_uid.' is on exclusion list!';
                $out_panic = $Plugins::MSAD::panic;
            }
            else {
                $search  = $ad_ldap->search(
                    base   => $cfg{'AD_base'},
                    filter => '(sAMAccountName=' . $del_uid . ')',
                    attrs  => ['distinguishedName'],
                    scope  => 'sub'
                );
                $code_del  = $search->code;
                $error_del = $search->error;
                $error_del =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun(looking for): uid=" . $del_uid . ";code=" . $code_del . ";error=" . $error_del)
                  if $config::aosi_debug_level > 2;
                if ($code_del == 0) {
                    @entries = $search->entries;
                    if ($#entries == -1) {
                        $code = -10031;
                        $msg  = "User $del_uid does not exists!";
                    }
                }
                else {
                    $msg  = 'Error searching for ' . $del_uid . ' (AD code: ' . $code_del . '; AD message: ' . $error_del . ').';
                    $code = -10030;
                }
            }
        }
    }
    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
# Delete an account
##################################################################
sub afterDeleteUser {
    my ($status_hash, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $ad_ldap, $mesg_ro, $code_ro, $error_ro);
    my ($req_dn, $code, $error, $req, $dec_pwd, $uid, $realm, $search);
    my ($new_uid, $new_dn, $code_new, $error_new, $temp,     $admin);
    my ($dn,      $del_dn, $attrs,    $ent_attrs, $mesg_del, $out_panic);
    my ($before_code, $before_msg, $before_panic, $msg);
    $this_fun = (caller(0))[3];
    $msg      = '';
    $code     = 0;
    ($uid, $realm) = split(/\@/, $user);

    AOSICFG::write_ws_log("|> $this_fun(in): user=$user;entry_in=$entry_in;out_xml=$out_xml;") if $config::aosi_debug_level > 1;

    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
    $before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
    $before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};

    if ($before_code == 0 and $before_panic == 0) {

        ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd, 1);
        if ($code == 0) {
            my $mesg = $ad_ldap->bind($cfg{'AD_aosiwrt_dn'}, password => $cfg{'AD_aosiwrt_pwd'});
            $code  = $mesg->code;
            $error = $mesg->error;
            $error =~ s/[\x00\n\r]//g;
            AOSICFG::write_ws_log("|> $this_fun(AOSIWRT bind): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
            if ($code == 0) {
                $dn  = $entry_in->{'dn'};
                $uid = get_uid($dn);
                ($del_dn, $code, $error) = get_ADDN($ad_ldap, $cfg{'AD_base'}, $uid);
                AOSICFG::write_ws_log("|> $this_fun: del_dn=" . $del_dn) if $config::aosi_debug_level > 1;
                if ($code == 0) {
                    $mesg_del = $ad_ldap->delete($del_dn);
                    $code     = $mesg_del->code;
                    $error    = $mesg_del->error;
                    $error =~ s/[\x00\n\r]//g;
                    AOSICFG::write_ws_log("|> $this_fun(deleting): code=" . $code . ";error=" . $error) if $config::aosi_debug_level > 2;
                    if ($code == 0) {
                        $msg  = 'User ' . $uid . ' deleted (AD code: ' . $code . '; AD message: ' . $error . ').';
                        $code = 0;
                    }
                    else {
                        $msg  = 'User ' . $uid . ' could not be deleted (AD code: ' . $code . '; AD message: ' . $error . ').';
                        $code = -10043;
                    }

                }
                else {
                    $msg  = 'DN for user ' . $uid . ' could not be found (AD code: ' . $code . '; AD message: ' . $error . ').';
                    $code = -10042;
                }
            }
            else {
                $msg  = 'AOSIWRT user could not bind to AD (AD code: ' . $code . '; AD message: ' . $error . ').';
                $code = -10041;
            }
        }
    }
    else {
        $code = -10040;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }
    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Check if adding an existig value or deleting a nonexisting value
##################################################################
# Same function for all actions
# 1 = beforeAddAttribute
# 2 = beforeModifyAttribute
# 3 = beforeDeleteAttribute
sub beforeChangeAttribute {
    my ($status_hash, $type, $ldap, $base_dn, $entry_in, $user, $pwd) = @_;
    my ($this_fun, $msg, $code,   $out_panic, $filter, $code_mod, $error_mod, $ad_ldap);
    my ($uid,      $dn,  $search, $att,       $val,    @entries,  $user_uid,  $realm);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;ldap=$ldap;base=$base_dn;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;

    ($code, $msg) = ck_filter($ldap, $base_dn, $entry_in, $user, $pwd, 0, $Plugins::MSAD::filters);
    if ($code == 0) {
        ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd);
        if ($code == 0) {
            $dn  = $entry_in->{'dn'};
            $att = $entry_in->{'attrs'}->[0]->{'att'};
            $val = $entry_in->{'attrs'}->[0]->{'val'}->[0];
            AOSICFG::write_ws_log("|> $this_fun: att=$att;val=$val") if ($att ne 'userPassword') and ($config::aosi_debug_level > 2);
            $uid = get_uid($dn);
            ($user_uid, $realm) = split(/\@/, $user);
            if (($uid eq $user_uid) or (user_is_ADadmin($user) > 0)) {
                AOSICFG::write_ws_log('|> ' . $this_fun . ': uid=' . $uid . ';user=' . $user . ';user_id=' . $user_uid) if $config::aosi_debug_level > 2;
                $filter = '(&(sAMAccountName=' . $uid . ')';
                if($att ne 'userPassword'){
                    $filter .= '(' . $att . '=' . $val . ')';
                }
                $filter .= ')';
                $search = $ad_ldap->search(
                    base   => $cfg{'AD_base'},
                    filter => $filter,
                    attrs  => [$att],
                    scope  => 'sub'
                );
                $code_mod  = $search->code;
                $error_mod = $search->error;
                $error_mod =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun(looking for): filter=$filter;code=" . $code_mod . ";error=" . $error_mod) if $config::aosi_debug_level > 2;
                if ($code_mod == 0) {
                    $msg = encode_base64($val);
                    @entries = $search->entries;
                    AOSICFG::write_ws_log("|> $this_fun(found): no_of_entries=" . $#entries) if $config::aosi_debug_level > 2;

                    # Adding an existing value
                    if (($type == 1) and ($#entries > -1)) {
                        $code      = -10051;
                        $msg       = "Attribute $att with value $val already exists!";
                        $out_panic = $Plugins::MSAD::panic;
                    }

                    # Deleting a nonexisting value
                    if (($type == 3) and ($#entries == -1)) {
                        $code      = -10052;
                        $msg       = "There is no attribute $att with value $val!";
                        $out_panic = $Plugins::MSAD::panic;
                    }
                }
            }
            else {
                $code = -10050;
                $msg  = 'User can change only his attributes!';
                $out_panic = $Plugins::MSAD::panic;
            }
        }
    }
    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

##################################################################
#  Change attributes (userPassword in AD is in fact unicodePwd)
##################################################################
# Same function for all actions
# 1 = afterAddAttribute
# 2 = afterModifyAttribute
# 3 = afterDeleteAttribute
sub afterChangeAttribute {
    my ($status_hash, $type, $entry_in, $out_xml, $user, $pwd) = @_;
    my ($this_fun, $out_panic, $msg, $mesg_mod, $mod_dn, $dec_pwd, $code);
    my ($uid, $dn, $att, $val, $code_mod, $error_mod, @entries, $user_uid, $realm);
    my ($before_code, $before_msg, $before_panic, $ad_ldap, $nval, $error);
    $this_fun  = (caller(0))[3];
    $msg       = '';
    $code      = 0;
    $out_panic = 0;

    AOSICFG::write_ws_log("|> $this_fun(in): tpye=$type;user=$user;entry_in=$entry_in;") if $config::aosi_debug_level > 1;
    ($code, $msg) = Plugins::parse_ldap_xml($out_xml);
    AOSICFG::write_ws_log("|> $this_fun: ldap_code=$code;ldap_msg=$msg;") if $config::aosi_debug_level > 2;

    $before_code  = $status_hash->{$plugin_name}->{'before'}->{'code'};
    $before_msg   = $status_hash->{$plugin_name}->{'before'}->{'msg'};
    $before_panic = $status_hash->{$plugin_name}->{'before'}->{'panic'};
    if ($before_code == 0 and $before_panic == 0) {
        ($ad_ldap, $code, $msg, $out_panic) = AD_RO_connect($user, $pwd);
        if ($code == 0) {
            $dn  = $entry_in->{'dn'};
            $att = $entry_in->{'attrs'}->[0]->{'att'};
            $val = $entry_in->{'attrs'}->[0]->{'val'}->[0];
            AOSICFG::write_ws_log("|> $this_fun: att=$att;val=$val") if ($att ne 'userPassword') and ($config::aosi_debug_level > 2);
            $uid = get_uid($dn);
            ($user_uid, $realm) = split(/\@/, $user);
            if (($uid eq $user_uid) or (user_is_ADadmin($user) > 0)) {
                AOSICFG::write_ws_log('|> ' . $this_fun . ': uid=' . $uid . ';user=' . $user . ';user_id=' . $user_uid) if $config::aosi_debug_level > 2;
                if ($att eq 'userPassword') {
                    $att = 'unicodePwd';
                    $val = decode_base64($before_msg);
                    map { $nval .= "$_\000" } split(//, "\"$val\"");
                    $val = $nval;
                }
                AOSICFG::write_ws_log("|> $this_fun: bind_as_admin") if $config::aosi_debug_level > 2;
                $mesg_mod = $ad_ldap->bind($cfg{'AD_aosiwrt_dn'}, password => $cfg{'AD_aosiwrt_pwd'});
                $code     = $mesg_mod->code;
                $msg      = $mesg_mod->error;
                $msg      =~ s/[\x00\n\r]//g;
                AOSICFG::write_ws_log("|> $this_fun(bind): code=" . $code . ";error=" . $msg) if $config::aosi_debug_level > 2;
                if ($code == 0) {
                    ($mod_dn, $code, $error) = get_ADDN($ad_ldap, $cfg{'AD_base'}, $uid);
                    AOSICFG::write_ws_log("|> $this_fun(replacing): $att=$val for dn=$mod_dn") if $att ne 'unicodePwd';
                    if ($type == 1 or $type == 2) {
                        $mesg_mod = $ad_ldap->modify($mod_dn, replace => { $att => $val });
                    }
                    elsif ($type == 3) {
                        $mesg_mod = $ad_ldap->modify($mod_dn, delete => { $att => $val });
                    }
                    $code = $mesg_mod->code;
                    $msg  = $mesg_mod->error;
                    $msg =~ s/[\x00\n\r]//g;
                }
            }
            else {
                $code = -10061;
                $msg  = 'User can change only his attributes!';
            }
        }
    }
    else {
        $code = -10060;
        $msg  = "Before action returned: panic=" . $before_panic . "; code=" . $before_code . "; message=" . $before_msg;
    }

    AOSICFG::write_ws_log('|> ' . $this_fun . '(out): code=' . $code . ';msg=' . $msg) if $config::aosi_debug_level > 1;
    return $code, $msg, $out_panic;
}

1;
